/*
Copyright 2025 BewitchedFlora
SPDX-License-Identifier: Apache-2.0
*/

#include "/lib/code/get_view_position.glsl"
#include "/lib/code/options.glsl"

uniform mat4 dhProjection;
uniform sampler2D gtexture;
uniform sampler2D noisetex;
uniform float viewWidth;

in vec2 texCoord;

layout(location = 0) out vec4 pixelColor;

const int AO_MAX_KERNEL_SIZE = 8;
const float AO_KERNEL_SCALE = 1.0 / float(AO_MAX_KERNEL_SIZE);
const float AO_SAMPLE_RADIUS = 5.0;
const vec2 AO_NOISE_SCALE = vec2(1280, 1280);
const vec3 AO_KERNEL[AO_MAX_KERNEL_SIZE] = vec3[](vec3(-0.007, -0.007, 0.014),vec3(-0.035, 0.054, 0.064),vec3(-0.028, -0.026, 0.028),vec3(-0.005, -0.053, 0.008), vec3(0.012, -0.028, 0.051),vec3(-0.065, 0.004, 0.015),vec3(0.032, 0.015, 0.006),vec3(-0.026, 0.063, 0.021));
const vec2 BAO_KERNEL[AO_MAX_KERNEL_SIZE] = vec2[](vec2(1.0, 0.0), vec2(-1.0, 1.0), vec2(-1.0, -1.0), vec2(0.0, -1.0), vec2(1.0, -1), vec2(-1.0, 0.0), vec2(0.0, 1.0), vec2(1.0, 1.0));

void main() {
    float screenPixelSize = 1.0 / viewWidth * 2.0;

    vec3 viewPos = get_view_position(texCoord);
    vec3 viewNormal = normalize(cross(dFdy(viewPos), dFdx(viewPos)) * -1.0);
    vec3 random = texture(noisetex, texCoord * AO_NOISE_SCALE).xyz; 
    vec3 tangent = normalize(random - viewNormal * dot(random, viewNormal));
    vec3 bitangent = cross(viewNormal, tangent);
    mat3 TBN = mat3(tangent, bitangent, viewNormal); 
    float viewDepth = viewPos.z + 0.001;

    float occlusion = 0.0;
    
    for (int i = 0; i < AO_MAX_KERNEL_SIZE; i++) {
        // SSAO Pass 1
        vec3 sampleCoord = viewPos + TBN * AO_KERNEL[i] * AO_SAMPLE_RADIUS;
        vec4 sampleOffset = dhProjection * vec4(sampleCoord, 1.0);
        sampleOffset.xy = (sampleOffset.xy / sampleOffset.w) * 0.5 + 0.5;
        float sampleDepth = get_view_position(sampleOffset.xy).z;
        float rangeCheck = smoothstep(0.0, 1.0, 1 / abs(viewPos.z - sampleDepth));
       
        // SSAO Pass 2
        vec2 boaSampleCoord = texCoord + BAO_KERNEL[i]*screenPixelSize;
        float boaSampleDepth = get_view_position(boaSampleCoord).z + 0.001;
        float rangeFalloff = smoothstep(-50.0, 0, boaSampleDepth);

        occlusion += ( (boaSampleDepth > viewDepth ? rangeFalloff * 1.2 : 0.0) + (float(sampleDepth >= sampleCoord.z + 0.0001) ) * rangeCheck * smoothstep(-30.0, 0.0, sampleDepth) );
    }

    vec4 texColor = texture(gtexture, texCoord);
    float occlusionFactor = occlusion * AO_KERNEL_SCALE;
    pixelColor = texColor * min(pow(1.2 - occlusionFactor / 4, 4), 1);
}