/*
Copyright 2025 BewitchedFlora
SPDX-License-Identifier: Apache-2.0
*/

#include "/lib/code/options.glsl"

in vec2 texCoord;
in vec2 lightCoord;
in vec3 normal;
in vec3 vertexPosition;
in float cloudiness;
in float vertexDistance;
in float reflectiveness;
in vec4 vertexColor;

uniform sampler2D gtexture;
uniform sampler2D lightmap;
uniform sampler2D gaux1;

uniform vec3 cameraPosition;
uniform float rainStrength;
uniform float fogStart;
uniform float fogEnd;
uniform vec3 fogColor;
uniform vec3 shadowLightPosition;

uniform vec3 sunPosition;
uniform int isEyeInWater;
uniform int worldTime;
uniform float frameTimeCounter;

layout(location = 0) out vec4 pixelColor;

void main() {
	vec3 camPos = vertexPosition.xyz + cameraPosition.xyz;
	
	vec4 texColor = texture(gtexture, texCoord);
	if (texColor.a < 0.1) discard;

	// Light and fog calculations
	vec4 lightColor = texture(lightmap, lightCoord);
	vec4 lightGradientTint = mix(vec4(0.45, 0.45, 0.75, 0.5), vec4(1.0, 1.0, 0.9, 1.0), lightColor.r - 0.1);
	lightColor = mix(lightColor, lightGradientTint, 0.5);

	// Calculate the warmth of light. Night has bright, warmer lights than day.
	float worldTimeBrightness = smoothstep(worldTime < 18000.0 ? 0.0 : 23992.0, 18000.0, worldTime) * 2.0 - 1.0;
	lightColor *= vec4(1, 1, 1, 1) + vec4(0.3, 0.15, 0.02, 1) * max(worldTimeBrightness, 0);
	
	// Make water invisible when you're near it.
	float normalizedDistance = clamp((vertexDistance - 0.5) / (4 - 1), 0.1, 0.9);
	lightColor.a = 1 - (1 - (normalizedDistance))*reflectiveness;

	// Matte color prior to material post-effects
	vec4 finalColor = texColor * vertexColor * vertexColor * lightColor * lightColor;
	finalColor *= vec4(1.05, 1.05, 1.05, 1);
	#ifdef SHADER_TERRAIN
		finalColor *= vec4(1.1, 1.1, 1.1, 1);

		float dimensionCloudiness = cloudiness/2;
		// Water surface texture effects
		#ifdef DIMENSION_NETHER
			#ifdef NETHER_LIGHT_CAUSTICS
				dimensionCloudiness = 0.5;
			#endif
		#endif
	
		#ifdef WATER_SURFACE_FOG
			vec2 waterCoord = vec2(frameTimeCounter / 50) - (camPos.x + camPos.zy) / 5; 
			vec3 surfaceRipple = texture(gaux1, waterCoord).xyz + texture(gaux1, -waterCoord).xyz;
			finalColor.rgb *= 1.0 + surfaceRipple * dimensionCloudiness;
			finalColor.a -= (1.0 - clamp(abs(camPos.x) + abs(camPos.y) + abs(camPos.z) / 5.0 + 0.1, 0.0, 1.0)) * reflectiveness;
   			finalColor.rgb *= 1.0 + fogColor * cloudiness * reflectiveness / 2.0;
		#endif

		// Underwater caustic texture effects
		#ifdef WATER_CAUSTICS
			vec2 waterCausticCoord = waterCoord;
			finalColor.rgb *= 1.0 + (texture(gaux1, waterCausticCoord).xyz * isEyeInWater + cloudiness / 4.0);
		#endif

		// Specular highlights
		vec3 lightDir = normalize(vec3((vec4(sunPosition, 1.0) * gl_ModelViewMatrix).xyz) - vertexPosition);
		float specular = max(dot(lightDir, reflect(-lightDir, normal)), 0.0);
		specular = pow(specular, 16.0) / 4.0; 
		finalColor.rgb += fogColor * min(specular * lightColor.r / 2.0, 1.0);
	#endif

	float fogVisibility = min(smoothstep(fogStart, fogEnd, vertexDistance * 2) + vertexDistance / 100.0 + rainStrength / 2.0, 1.0);
	fogVisibility = (isEyeInWater == 1) ? fogVisibility/1.2 : fogVisibility;
	#ifdef DIMENSION_NETHER
		fogVisibility = min(smoothstep(fogStart, fogEnd, vertexDistance/6) + vertexDistance / 100.0, 1.0);
	#endif
	
	pixelColor = vec4(mix(finalColor.rgb, fogColor, fogVisibility), finalColor.a);
}